@@ -28,10 +28,16 @@
 
 #ifdef HAVE_WX
 #define getc getFromWX_2
+#define ungetc wxUnget_c
+int reading_char_now = 0;
 #endif
 
 #ifdef HAVE_TERMIO_H
+#ifdef HAVE_WX
+#include <termios.h>
+#else
 #include <termio.h>
+#endif
 #else
 #ifdef HAVE_SGTTY_H
 #include <sgtty.h>
@@ -60,6 +66,8 @@ extern int getch(), kbhit();
 
 NODE *file_list = NULL;
 NODE *reader_name = NIL, *writer_name = NIL, *file_prefix = NIL;
+int need_save = 0;  /* nonzero if workspace changed since save */
+NODE *save_name = NIL;	/* filename of last save or load command */
 
 char *asciiz(NODE *arg) {
     char *out = (char *)malloc(getstrlen(arg)+1);
@@ -136,6 +144,8 @@ NODE *lprefix(NODE *args) {
 FILE *open_file(NODE *arg, char *access) {
     char *fnstr;
     FILE *tstrm;
+    char *old_stringptr = print_stringptr;
+    int old_stringlen = print_stringlen;
 
     if (is_list(arg)) { /* print to string */
 	if (*access != 'w') {
@@ -163,12 +173,16 @@ FILE *open_file(NODE *arg, char *access) {
 	fnstr = (char *) malloc((size_t)getstrlen(arg) + 1);
     if (fnstr == NULL) {
 	err_logo(FILE_ERROR, make_static_strnode(message_texts[MEM_LOW]));
+	print_stringptr = old_stringptr;
+	print_stringlen = old_stringlen;
 	return NULL;
     }
     if (file_prefix != NIL) {
 	print_stringptr = fnstr;
 	ndprintf((FILE *)NULL, "%p%t%p", file_prefix, separator, arg);
 	*print_stringptr = '\0';
+	print_stringptr = old_stringptr;
+	print_stringlen = old_stringlen;
     } else
 	noparity_strnzcpy(fnstr, getstrptr(arg), getstrlen(arg));
     tstrm = fopen(fnstr, access);
@@ -370,16 +384,23 @@ NODE *lerasefile(NODE *arg) {
 
 NODE *lsave(NODE *arg) {
     FILE *tmp;
+    NODE *name;
 
+    if (arg == NIL) name = save_name;
+    else name = car(arg);
     tmp = writestream;
-    writestream = open_file(car(arg), "w+");
-    if (writestream != NULL) {
-	setcar(arg, cons(lcontents(NIL), NIL));
-	lpo(car(arg));
-	fclose(writestream);
-    }
-    else
-	err_logo(CANT_OPEN_ERROR, car(arg));
+    if (name != NIL) {
+	writestream = open_file(name, "w+");
+	if (writestream != NULL) {
+	    lpo(cons(lcontents(NIL), NIL));
+	    fclose(writestream);
+	    need_save = 0;
+	    save_name = name;
+	}
+	else
+	    err_logo(CANT_OPEN_ERROR, car(arg));
+    } else
+	err_logo(NOT_ENOUGH, NIL);
     writestream = tmp;
     return(UNBOUND);
 }
@@ -420,18 +441,20 @@ void silent_load(NODE *arg, char *prefix) {
 	arg = cnv_node_to_strnode(arg);
 	if (arg == UNBOUND) return;
 	if (!strncmp(getstrptr(arg), ".", getstrlen(arg))) return;
+	if (!strncmp(getstrptr(arg), "..", 2)) return;
+	if (getstrlen(arg) > 150) return;
 	noparitylow_strnzcpy(load_path + (int)strlen(load_path),
 			     getstrptr(arg), getstrlen(arg));
 	if (prefix == NULL)
 	  strcat(load_path, ".lg");
-#ifdef WIN32
+/* #ifdef WIN32 */
 	else if (arg != NIL) {
 	    char *cp;
 	    for (cp = load_path; *cp != '\0'; cp++)
 		if (*cp == '?') *cp = 'Q';
 	}
 	/*  strcpy(load_path, eight_dot_three(load_path));  */
-#endif	
+/* #endif */
     }
     tmp_stream = loadstream;
     tmp_line = current_line;
@@ -457,6 +480,7 @@ NODE *lcslsload(NODE *arg) {
     file_prefix = NIL;
     silent_load(car(arg),csls);
     file_prefix = save_prefix;
+    save_name = car(arg);
     return UNBOUND;
 }
 
@@ -475,6 +499,7 @@ NODE *lload(NODE *arg) {
 	    if (exec_list != NIL) eval_driver(exec_list);
 	}
 	fclose(loadstream);
+	save_name = car(arg);
 	runstartup(st);
     } else
 	err_logo(CANT_OPEN_ERROR, car(arg));
@@ -513,14 +538,31 @@ NODE *lreadrawline(NODE *args) {
     return(val);
 }
 
+int readchar_lookahead_buf = -1;
+
 NODE *lreadchar(NODE *args) {
 #ifdef WIN32
     MSG msg;
 #endif /* WIN32 */
     char c;
 
+    if (readchar_lookahead_buf >= 0) {
+	c = (char)readchar_lookahead_buf;
+	readchar_lookahead_buf = -1;
+	return(make_strnode((char *)&c, (struct string_block *)NULL, 1,
+		(getparity(c) ? BACKSLASH_STRING : STRING), strnzcpy));
+    }
     charmode_on();
     input_blocking++;
+#ifdef HAVE_WX
+	if (interactive && readstream==stdin) {
+	    reading_char_now = 1;
+	    c = getFromWX();
+	    reading_char_now = 0;
+	}
+	else
+	    c = (char)getc(readstream);
+#else
 #ifndef TIOCSTI
     if (!setjmp(iblk_buf))
 #endif
@@ -570,8 +612,13 @@ NODE *lreadchar(NODE *args) {
 #endif /* ibm */
 #endif /* mac */
     }
+#endif /* wx */
     input_blocking = 0;
+#ifdef HAVE_WX
+    if ((!interactive || readstream!=stdin) && feof(readstream)) {
+#else
     if (feof(readstream)) {
+#endif
 	return(NIL);
     }
     return(make_strnode((char *)&c, (struct string_block *)NULL, 1,
@@ -583,6 +630,7 @@ NODE *lreadchars(NODE *args) {
     struct string_block *strhead = NULL;
     char *strptr= NULL;
     NODETYPES type = STRING;
+    char *cp;
 
     c = (unsigned int)getint(pos_int_arg(args));
     if (stopping_flag == THROWING) return UNBOUND;
@@ -598,7 +646,19 @@ NODE *lreadchars(NODE *args) {
 	    return UNBOUND;
 	}
 	strptr = strhead->str_str;
+#ifdef HAVE_WX
+	if (interactive && readstream==stdin) {
+	    reading_char_now = 1;
+	    cp=strptr;
+	    for (i=c; i != 0; --i) {
+		*cp++ = getFromWX();
+	    }
+	    reading_char_now = 0;
+	} else
+	    c = (unsigned int)fread(strptr, 1, (size_t)c, readstream);
+#else
 	c = (unsigned int)fread(strptr, 1, (size_t)c, readstream);
+#endif
 	setstrrefcnt(strhead, 0);
     }
     input_blocking = 0;
@@ -617,6 +677,9 @@ NODE *lreadchars(NODE *args) {
 NODE *leofp(NODE *args) {
     int c;
 
+#ifdef HAVE_WX
+    if (interactive && readstream==stdin) return FalseName();
+#endif
     c = getc(readstream);
     if (c == EOF) return(TrueName());
 
@@ -625,7 +688,7 @@ NODE *leofp(NODE *args) {
 }
 
 NODE *lkeyp(NODE *args) {
-#ifdef unix
+#if defined(unix) | defined(__WXMSW__)
     long nc;
 #endif
     int c;
@@ -634,9 +697,14 @@ NODE *lkeyp(NODE *args) {
     int old_mode;
 #endif
 
+    if (readchar_lookahead_buf >= 0) return TrueName();
     if (readstream == stdin && interactive) {
 	charmode_on();
 	fflush(stdout);
+	fix_turtle_shownness();
+#ifdef HAVE_WX
+	return(wxKeyp() ? TrueName() : FalseName());
+#else
 #if defined(__RZTC__) && !defined(WIN32) /* sowings */
 	zflush();
 #endif
@@ -668,7 +736,7 @@ NODE *lkeyp(NODE *args) {
 #endif
 #else
 #ifdef FIONREAD
-	ioctl(0,FIONREAD,(char *)(&nc));
+	ioctl(0,FIONREAD,(int *)(&nc));
 #else
 	ndprintf(stdout, "%t\n", message_texts[NO_FIONREAD]);
 	nc = 1;    /* pretend there's a char so we don't loop */
@@ -678,6 +746,7 @@ NODE *lkeyp(NODE *args) {
 	else
 	    return(FalseName());
 #endif
+#endif /* wx */
     }
     c = getc(readstream);
     if (feof(readstream))
